import re

## this module
from .dumpfmt import J2substDumpFormat
from .defaults import (
    J2SUBST_DUMP_FORMAT,
    J2SUBST_EMPTY_JSON,
    J2SUBST_EMPTY_YAML,
    J2SUBST_MAX_DEPTH,
)
from .functions import (
    re_sub,
)


J2SUBST_CLI_HELP__DUMP = f'''
By default, dump format is {J2SUBST_DUMP_FORMAT.name}.

Available formats: {', '.join([e.name for e in J2substDumpFormat])}.

Format name is case-insensitive.

Empty config in YAML format is equivalent to:

{re_sub(J2SUBST_EMPTY_YAML.strip(), '^', ' ' * 2, opt = re.MULTILINE)}

Empty config in JSON format is equivalent to:

{re_sub(J2SUBST_EMPTY_JSON.strip(), '^', ' ' * 2, opt = re.MULTILINE)}
'''


J2SUBST_CLI_HELP__TEMPLATE_PATH = '''
Template path is colon-separated list of paths to search for templates.

Path component may contain placeholders:

@{CWD}    - current working directory
@{ORIGIN} - "origin" path

From CLI side, "origin" is always directory where template file is located.
From API side, "origin" maybe specified as any arbitrary directory.

If path component contains placeholder, it is replaced with the corresponding value.

If path component is still containing placeholder prefix ("@{"), it is skipped.

If path component does not contain [unresolved] placeholders, it is treated as directory.

'''


J2SUBST_CLI_HELP__ENV = '''
Corresponding environment variables are also supported.

|------------------------+------------------+---------|
| Environment variable   | Flag option      | Type    |
|------------------------+------------------+---------|
| J2SUBST_VERBOSE        | --verbose        | integer |
| J2SUBST_QUIET          | --quiet          | flag    |
| J2SUBST_DEBUG          | --debug          | flag    |
| J2SUBST_STRICT         | --strict         | flag    |
| J2SUBST_FORCE          | --force          | flag    |
| J2SUBST_UNLINK         | --unlink         | flag    |
| J2SUBST_DEPTH          | --depth          | integer |
| J2SUBST_CONFIG_PATH    | --config-path    | string  |
| J2SUBST_TEMPLATE_PATH  | --template-path  | string  |
| J2SUBST_PYTHON_MODULES | --python-modules | string  |
| J2SUBST_FN_FILTERS     | --fn-filters     | flag    |
| J2SUBST_DICT_NAME_CFG  | --dict-name-cfg  | string  |
| J2SUBST_DICT_NAME_ENV  | --dict-name-env  | string  |
|------------------------+------------------+---------|

See "--help-click" for more details about how Click handles environment variables, especially for flag options.

See "--help-cicd" for more details about how J2subst handles default values while in CI/CD.
'''


J2SUBST_CLI_HELP__CLICK = '''
Variable names are:
- Case-insensitive on Windows but not on other platforms.
- Not stripped of whitespaces and should match the exact name provided to the argument.

For flag options, there is two concepts to consider: the activation of the flag driven by the environment variable, and the value of the flag if it is activated.

The environment variable need to be interpreted, because values read from them are always strings.
We need to transform these strings into boolean values that will determine if the flag is activated or not.

Here are the rules used to parse environment variable values for flag options:
- "true", "1", "yes", "on", "t", "y" are interpreted as activating the flag.
- "false", "0", "no", "off", "f", "n" are interpreted as deactivating the flag.
- The presence of the environment variable without value is interpreted as deactivating the flag.
- Empty strings are interpreted as deactivating the flag.
- Values are case-insensitive, so the "True", "TRUE", "tRuE" strings are all activating the flag.
- Values are stripped of leading and trailing whitespaces before being interpreted, so the " True " string is transformed to "true" and so activates the flag.
- Any other value is interpreted as deactivating the flag.

---

Ref: https://click.palletsprojects.com/en/stable/options/#values-from-environment-variables
'''

J2SUBST_CLI_HELP__CICD = f'''
CI/CD notes:

- option "--depth" / variable "J2SUBST_DEPTH" defaults to "{J2SUBST_MAX_DEPTH}" if running in CI/CD and "1" otherwise.

- if argument list is empty then it set to current working directory.
'''



import hashlib
import io
import os
import os.path
import re
import sys

from collections.abc import (
    Mapping,
    Sequence,
)
from os import (
    PathLike,
)
from typing import (
    Any,
    Callable,
)

## this module
from .defaults import (
    J2SUBST_ENV_CI,
    J2SUBST_ENV_SKIP,
)


def is_str(x: Any) -> bool:
    return isinstance(x, str)


def is_str_or_path(x: Any) -> bool:
    return is_str(x) or isinstance(x, PathLike)


def is_seq(x: Any) -> bool:
    return isinstance(x, Sequence) and not is_str_or_path(x)


def is_map(x: Any) -> bool:
    return isinstance(x, Mapping)


def is_plain_key(x: Any | None) -> bool:
    if not (is_str_or_path(x) and x):
        return False
    return bool(re.fullmatch(r'[a-zA-Z_][a-zA-Z0-9_]*', str(x)))


def uniq(a: Sequence[Any]) -> list[Any]:
    return list(set(a))


def only_str(a: Sequence[Any]) -> list[str]:
    return [str(x) for x in filter(is_str_or_path, a)]


def non_empty_str(a: Sequence[Any]) -> list[str]:
    return only_str(list(filter(None, a)))


def uniq_str_list(a: Sequence[Any]) -> list[Any]:
    return uniq(non_empty_str(a))


def str_split_to_list(s: str, sep: str | re.Pattern[str] = r'\s+') -> list[str]:
    return non_empty_str(re.split(sep, s))


def dict_to_str_list(x: dict[str, Any]) -> list[str]:
    r: list[str] = []
    for k in sorted(x.keys()):
        if x[k] is None:
            r.append(f'{k}')
        else:
            r.append(f'{k}={str(x[k])}')
    return r


def any_to_str_list(x: Any) -> list[str]:
    if x is None:
        return []
    if is_str_or_path(x):
        return [str(x)]
    if is_seq(x):
        return [str(e) for e in x]
    if is_map(x):
        return dict_to_str_list(x)
    return [str(x)]


def is_re_match(x: Any, pat: str | re.Pattern[str], opt: int = 0) -> bool:
    if is_str_or_path(x):
        return bool(re.match(pat, str(x), opt))
    if is_seq(x):
        return any(is_re_match(v, pat, opt) for v in x)
    if is_map(x):
        return any(is_re_match(v, pat, opt) for v in x.keys())
    return False


def is_re_fullmatch(x: Any, pat: str | re.Pattern[str], opt: int = 0) -> bool:
    if is_str_or_path(x):
        return bool(re.fullmatch(pat, str(x), opt))
    if is_seq(x):
        return any(is_re_fullmatch(v, pat, opt) for v in x)
    if is_map(x):
        return any(is_re_fullmatch(v, pat, opt) for v in x.keys())
    return False


def re_match(x: Any, pat: str | re.Pattern[str], opt: int = 0) -> Any:
    if is_str_or_path(x):
        return re.match(pat, str(x), opt)
    if is_seq(x):
        return [v for v in x if re_match(v, pat, opt)]
    if is_map(x):
        return {k: v for k, v in x.items() if re_match(k, pat, opt)}
    return None


def re_fullmatch(x: Any, pat: str | re.Pattern[str], opt: int = 0) -> Any:
    if is_str_or_path(x):
        return re.fullmatch(pat, str(x), opt)
    if is_seq(x):
        return [v for v in x if re_fullmatch(v, pat, opt)]
    if is_map(x):
        return {k: v for k, v in x.items() if re_fullmatch(k, pat, opt)}
    return None


def re_match_neg(x: Any, pat: str | re.Pattern[str], opt: int = 0) -> Any:
    if is_str_or_path(x):
        return not bool(re.match(pat, str(x), opt))
    if is_seq(x):
        return [v for v in x if re_match_neg(v, pat, opt)]
    if is_map(x):
        return {k: v for k, v in x.items() if re_match_neg(k, pat, opt)}
    return x


def re_fullmatch_neg(x: Any, pat: str | re.Pattern[str], opt: int = 0) -> Any:
    if is_str_or_path(x):
        return not bool(re.fullmatch(pat, str(x), opt))
    if is_seq(x):
        return [v for v in x if re_fullmatch_neg(v, pat, opt)]
    if is_map(x):
        return {k: v for k, v in x.items() if re_fullmatch_neg(k, pat, opt)}
    return x


def dict_remap_keys(x: dict[Any, Any], key_map: Callable[[Any], Any] | None) -> dict[Any, Any]:
    if key_map is None:
        print('J2subst: dict_remap_keys(): key_map is None', file=sys.stderr)
        return x
    m: dict[Any, Any] = {}
    for k in x.keys():
        v = key_map(k)
        if v in m:
            ## merely debug output
            print(f'J2subst: dict_remap_keys(): duplicate key {repr(v)} <= {repr(k)}', file=sys.stderr)
            continue
        m[v] = x[k]
    return m


def re_sub(x: Any, pat: str | re.Pattern[str], repl: str | Callable[[re.Match[str]], str], count: int = 0, opt: int = 0) -> Any:
    if is_str_or_path(x):
        return re.sub(pat, repl, str(x), count, opt)
    if is_seq(x):
        return [re_sub(v, pat, repl, count, opt) for v in x]
    if is_map(x):
        return dict_remap_keys(x, lambda k: re_sub(k, pat, repl, count, opt))
    return x


def any_to_env_dict(x: Any) -> dict[str, str | None]:
    if x is None:
        return {}

    h: dict[str, str | None] = {}

    ## TODO: review
    def __feed_dict(key: Any, value: Any = None, parse_key: bool = True):
        if not key:
            return
        if (value is None) and (not parse_key):
            return
        k = str(key)
        if not k:
            return
        v = value
        if parse_key:
            k2, m, v2 = k.partition('=')
            if m == '=':
                k = k2
                v = v2
        if is_env_skipped(k) or (not is_plain_key(k)):
            return
        if k in h:
            return
        if v is not None:
            v = str(v)
        h[k] = v

    if is_str_or_path(x):
        __feed_dict(str(x))
    elif is_seq(x):
        for e in x:
            __feed_dict(e)
    elif is_map(x):
        for k in x:
            __feed_dict(k, x[k], False)
    else:
        return {}

    return h


def dict_keys(x: dict[Any, Any]) -> list[Any]:
    return sorted(x.keys())


def dict_empty_keys(x: dict[Any, Any]) -> list[Any]:
    return sorted([k for k in x.keys() if x[k] is None])


def dict_non_empty_keys(x: dict[Any, Any]) -> list[Any]:
    return sorted([k for k in x.keys() if x[k] is not None])


def list_diff(a: Sequence[Any], b: Sequence[Any]) -> list[Any]:
    return list(set(a) - set(b))


def list_intersect(a: Sequence[Any], b: Sequence[Any]) -> list[Any]:
    return list(set(a) & set(b))


## ref: https://click.palletsprojects.com/en/stable/options/#values-from-environment-variables
def click_bool(x: Any) -> bool:
    if is_str(x):
        return str(x).strip().lower() in { "1", "on", "t", "true", "y", "yes" }
    if isinstance(x, bool):
        return x
    return False

## ref: https://click.palletsprojects.com/en/stable/options/#values-from-environment-variables
def click_bool_neg(x: Any) -> bool:
    if is_str(x):
        return str(x).strip().lower() in { "0", "f", "false", "n", "no", "off" }
    if isinstance(x, bool):
        return not x
    return False


## ref: https://pkg.go.dev/strconv#ParseBool
## behavior is mostly the same except no errors for invalid values
def go_bool(x: Any) -> bool:
    if is_str(x):
        return str(x) in { "1", "T", "TRUE", "True", "t", "true" }
    if isinstance(x, bool):
        return x
    return False

## ref: https://pkg.go.dev/strconv#ParseBool
## behavior is mostly the same except no errors for invalid values
def go_bool_neg(x: Any) -> bool:
    if is_str(x):
        return str(x) in { "0", "F", "FALSE", "False", "f", "false" }
    if isinstance(x, bool):
        return not x
    return False


## NB: not in J2SUBST_FUNCTIONS
def merge_dict_recurse(d1: dict[Any, Any] | None, d2: dict[Any, Any] | None) -> dict[Any, Any]:
    x: dict[Any, Any] = {}
    if d1:
        x = x | d1
    if not d2:
        return x

    keys1 = set(x.keys())
    keys2 = set(d2.keys())
    common = keys1 & keys2
    missing = keys2 - common

    map1 = {k for k in common if is_map(x.get(k))}
    seq1 = {k for k in common if is_seq(x.get(k))}
    misc1 = common - seq1 - map1

    merge_safe = missing | misc1
    x.update({k: d2.get(k) for k in merge_safe})

    map_common = {k for k in map1 if is_map(d2.get(k))}
    for k in map_common:
        y = d2.get(k)
        if not y:
            x[k] = {}
            continue
        x[k] = merge_dict_recurse(x.get(k), y)

    seq_common = {k for k in seq1 if is_seq(d2.get(k))}
    for k in seq_common:
        y = d2.get(k)
        if not y:
            x[k] = []
            continue
        ## TYPING-TODO
        x[k] = uniq(list(x.get(k)) + list(y))

    unmerged = (map1 - map_common) | (seq1 - seq_common)
    for k in unmerged:
        t1 = type(x.get(k))
        t2 = type(d2.get(k))
        print(f'J2subst: merge_dict_recurse(): skipping key {k} due to type mismatch: {t1} vs. {t2}', file=sys.stderr)

    return x


def join_prefix(prefix: str, *paths: Any) -> str:
    pfx = prefix or '/'
    pfx = '/' + pfx.strip('/')
    rv = os.path.normpath(os.path.join(pfx, *paths).rstrip('/')).rstrip('/')
    if rv == pfx:
        raise ValueError('J2subst: join_prefix(): empty path after prefix')
    common = os.path.commonpath([pfx, rv])
    if common == pfx:
        return rv
    ## slowpath
    rv = rv.removeprefix(common).lstrip('/')
    rv = os.path.join(pfx, rv)
    return rv


def md5(x: str) -> str:
    return hashlib.md5(x.encode('utf-8')).hexdigest()


def sha1(x: str) -> str:
    return hashlib.sha1(x.encode('utf-8')).hexdigest()


def sha256(x: str) -> str:
    return hashlib.sha256(x.encode('utf-8')).hexdigest()


def sha384(x: str) -> str:
    return hashlib.sha384(x.encode('utf-8')).hexdigest()


def sha512(x: str) -> str:
    return hashlib.sha512(x.encode('utf-8')).hexdigest()


def sha3_256(x: str) -> str:
    return hashlib.sha3_256(x.encode('utf-8')).hexdigest()


def sha3_384(x: str) -> str:
    return hashlib.sha3_384(x.encode('utf-8')).hexdigest()


def sha3_512(x: str) -> str:
    return hashlib.sha3_512(x.encode('utf-8')).hexdigest()


def file_md5(x: str) -> str:
    with open(x, 'rb') as f:
        return hashlib.md5(f.read()).hexdigest()


def file_sha1(x: str) -> str:
    with open(x, 'rb') as f:
        return hashlib.sha1(f.read()).hexdigest()


def file_sha256(x: str) -> str:
    with open(x, 'rb') as f:
        return hashlib.sha256(f.read()).hexdigest()


def file_sha384(x: str) -> str:
    with open(x, 'rb') as f:
        return hashlib.sha384(f.read()).hexdigest()


def file_sha512(x: str) -> str:
    with open(x, 'rb') as f:
        return hashlib.sha512(f.read()).hexdigest()


def file_sha3_256(x: str) -> str:
    with open(x, 'rb') as f:
        return hashlib.sha3_256(f.read()).hexdigest()


def file_sha3_384(x: str) -> str:
    with open(x, 'rb') as f:
        return hashlib.sha3_384(f.read()).hexdigest()


def file_sha3_512(x: str) -> str:
    with open(x, 'rb') as f:
        return hashlib.sha3_512(f.read()).hexdigest()


## NB: not in J2SUBST_FUNCTIONS
def is_env_skipped(x: Any) -> bool:
    if not is_str(x):
        return True
    x = str(x)
    for r in J2SUBST_ENV_SKIP:
        if re.match(r, x):
            return True
    return False


def j2subst_escape(x: Any) -> Any:
    if x is None:
        return None
    if is_str_or_path(x):
        x = str(x)
        if x == '':
            return "''"
        if re.search(r'(?:\s|[;{}()\[\]\\\'"*?])', x):
            return repr(x)
        return x
    if is_seq(x):
        return [ j2subst_escape(v) for v in x ]
    if is_map(x):
        return dict_remap_keys(x, j2subst_escape)
    return j2subst_escape(str(x))


def is_file_io(x: Any) -> bool:
    if not isinstance(x, io.IOBase):
        return False
    if x.closed:
        return False
    ## TODO: avoid try-except
    try:
        n = x.fileno()
        return n >= 0
    except (OSError, io.UnsupportedOperation):
        pass
    return False


def is_file_io_read(x: Any) -> bool:
    if not is_file_io(x):
        return False
    return x.readable()


def is_file_io_write(x: Any) -> bool:
    if not is_file_io(x):
        return False
    return x.writable()


__j2subst_stdin_fd: int | None = None
__j2subst_stdout_fd: int | None = None


def is_stdin(x: Any) -> bool:
    _STDIN = '/dev/stdin'

    if x is None:
        return False

    # pylint: disable=W0603
    global __j2subst_stdin_fd

    if is_file_io_read(x):
        if __j2subst_stdin_fd is None:
            if is_file_io_read(sys.stdin):
                __j2subst_stdin_fd = sys.stdin.fileno()
            else:
                __j2subst_stdin_fd = -1
        if __j2subst_stdin_fd is None:
            return False
        if __j2subst_stdin_fd < 0:
            return False

        return os.path.sameopenfile(x.fileno(), __j2subst_stdin_fd)

    if not (x and is_str_or_path(x)):
        return False
    if str(x) in [ '-', _STDIN ]:
        return True
    return os.path.exists(x) and os.path.samefile(x, _STDIN)


def is_stdout(x: Any) -> bool:
    _STDOUT = '/dev/stdout'

    if x is None:
        return False

    # pylint: disable=W0603
    global __j2subst_stdout_fd

    if is_file_io_write(x):
        if __j2subst_stdout_fd is None:
            if is_file_io_read(sys.stdout):
                __j2subst_stdout_fd = sys.stdout.fileno()
            else:
                __j2subst_stdout_fd = -1
        if __j2subst_stdout_fd is None:
            return False
        if __j2subst_stdout_fd < 0:
            return False

        return os.path.sameopenfile(x.fileno(), __j2subst_stdout_fd)

    if not (x and is_str_or_path(x)):
        return False
    if str(x) in [ '-', _STDOUT ]:
        return True
    return os.path.exists(x) and os.path.samefile(x, _STDOUT)


__j2subst_is_ci: bool | None = None


## NB: not in J2SUBST_FUNCTIONS
def is_ci(_x: Any = None) -> bool:
    # pylint: disable=W0603
    global __j2subst_is_ci

    if __j2subst_is_ci is not None:
        return __j2subst_is_ci

    ## ref: https://github.com/watson/ci-info/blob/3fae1ac492f59c1835a56b2a3c40b8c2cbeb02c1/index.js#L37

    # if os.getenv('CI') == 'false':
    if click_bool_neg(os.getenv('CI')):
        __j2subst_is_ci = False
    else:
        __j2subst_is_ci = False
        for e in list_intersect(J2SUBST_ENV_CI, list(os.environ.keys())):
            if os.environ[e]:
                __j2subst_is_ci = True
                break

    return __j2subst_is_ci


J2SUBST_FILTERS: list[Any] = [
    any_to_env_dict,
    any_to_str_list,
    click_bool,
    click_bool_neg,
    dict_empty_keys,
    dict_keys,
    dict_non_empty_keys,
    dict_remap_keys,
    dict_to_str_list,
    file_md5,
    file_sha1,
    file_sha256,
    file_sha384,
    file_sha512,
    file_sha3_256,
    file_sha3_384,
    file_sha3_512,
    go_bool,
    go_bool_neg,
    is_file_io,
    is_file_io_read,
    is_file_io_write,
    is_map,
    is_plain_key,
    is_re_fullmatch,
    is_re_match,
    is_seq,
    is_stdin,
    is_stdout,
    is_str,
    is_str_or_path,
    j2subst_escape,
    join_prefix,
    list_diff,
    list_intersect,
    md5,
    non_empty_str,
    only_str,
    re_fullmatch,
    re_fullmatch_neg,
    re_match,
    re_match_neg,
    re_sub,
    sha1,
    sha256,
    sha384,
    sha512,
    sha3_256,
    sha3_384,
    sha3_512,
    str_split_to_list,
    uniq,
    uniq_str_list,
]

J2SUBST_FILTER_ALIASES: dict[str, Any] = {
    'j2e': j2subst_escape,
}



from typing import (
    Any,
)

## this module
from .dumpfmt import J2substDumpFormat


J2SUBST_VERSION = '0.0.4'

J2SUBST_DICT_NAME_CFG = 'cfg'
J2SUBST_DICT_NAME_ENV = 'env'

J2SUBST_TEMPLATE_EXT = '.j2'

J2SUBST_TEMPLATE_PATH_PARTS = [ '@{ORIGIN}', '@{CWD}' ]
J2SUBST_TEMPLATE_PATH = ':'.join(J2SUBST_TEMPLATE_PATH_PARTS)

## merely ephemeral
J2SUBST_MAX_DEPTH = 20

## NB: leading dots are mandatory!
J2SUBST_CONFIG_EXT = [
    '.yaml', '.yml',
    '.json',
    '.toml',
]

J2SUBST_JINJA_EXTENSIONS = [
    'jinja2.ext.do',
    'jinja2.ext.loopcontrols',
]

J2SUBST_JINJA_DEBUG_EXTENSIONS = [
    'jinja2.ext.debug',
]

J2SUBST_PYTHON_MODULES = [
    'datetime',
    'hashlib',
    'pathlib',
    're',
    'secrets',
    'string',
]

J2SUBST_PYTHON_MODULE_ALIASES = {
    'os_path': 'os.path',
}

J2SUBST_BUILTIN_FUNCTIONS: list[Any] = [
    bool,
    # dict, ## already present
    filter,
    isinstance,
    len,
    list,
    repr,
    set,
    sorted,
    str,
    type,
]

J2SUBST_BUILTIN_FUNCTION_ALIASES: dict[str, Any] = {
}

J2SUBST_ENV_SKIP = [
    r'J2SUBST_',
    r'_$',
]

J2SUBST_ENV_CI = [
    'BUILD_ID',
    'BUILD_NUMBER',
    'CI',
    'CI_APP_ID',
    'CI_BUILD_ID',
    'CI_BUILD_NUMBER',
    'CI_NAME',
    'CONTINUOUS_INTEGRATION',
    'RUN_ID',
]

J2SUBST_DUMP_FORMAT = J2substDumpFormat.YAML

J2SUBST_EMPTY_YAML = '\n---\n# empty\n---\n'
J2SUBST_EMPTY_JSON = '{}'



import io
import os
import os.path
import sys
import importlib
import json
import tomllib

from collections.abc import (
    Mapping,
    Sequence,
)
from os import (
    PathLike,
)
from typing import (
    Any,
)

## jinja2
import jinja2
## natsort
import natsort
## pyyaml
import yaml

## this module
from .dumpfmt import J2substDumpFormat
from .defaults import (
    J2SUBST_BUILTIN_FUNCTION_ALIASES,
    J2SUBST_BUILTIN_FUNCTIONS,
    J2SUBST_CONFIG_EXT,
    J2SUBST_DICT_NAME_CFG,
    J2SUBST_DICT_NAME_ENV,
    J2SUBST_DUMP_FORMAT,
    J2SUBST_EMPTY_JSON,
    J2SUBST_EMPTY_YAML,
    J2SUBST_JINJA_DEBUG_EXTENSIONS,
    J2SUBST_JINJA_EXTENSIONS,
    J2SUBST_PYTHON_MODULE_ALIASES,
    J2SUBST_PYTHON_MODULES,
    J2SUBST_TEMPLATE_EXT,
    J2SUBST_TEMPLATE_PATH_PARTS,
)
from .functions import (
    J2SUBST_FILTERS,
    J2SUBST_FILTER_ALIASES,
    is_ci,
    is_env_skipped,
    is_map,
    is_plain_key,
    is_seq,
    is_stdin,
    is_stdout,
    merge_dict_recurse,
    non_empty_str,
)


class J2subst:

    def __init__(self,
                 dump_only: bool = False,

                 verbosity: int = 0,
                 debug: bool = False,
                 strict: bool = False,
                 force: bool = False,
                 unlink: bool = False,

                 config_path: Sequence[str | PathLike[str]] | None = None,
                 template_path: Sequence[str | PathLike[str]] | None = None,

                 python_modules: Sequence[str] | Mapping[str, str] | None = None,
                 filters_as_functions: bool = False,
                 dict_name_cfg: str = J2SUBST_DICT_NAME_CFG,
                 dict_name_env: str = J2SUBST_DICT_NAME_ENV,
    ):

        self.dump_only = bool(dump_only)
        if not self.dump_only:
            if not is_plain_key(dict_name_cfg):
                raise ValueError(f'not valid "dict_name_cfg": {repr(dict_name_cfg)}')
            self.dict_cfg_name = dict_name_cfg
            if not is_plain_key(dict_name_env):
                raise ValueError(f'not valid "dict_name_env": {repr(dict_name_env)}')
            self.dict_env_name = dict_name_env

        self.debug = bool(debug)
        self.verbosity = int(verbosity)

        self.force = bool(force)
        self.strict = bool(strict)
        self.unlink = False

        self.dict_cfg: dict[str, Any] = {}

        self.config_path: list[str] = []
        if config_path:
            self.config_path = non_empty_str(config_path)

        self.__merge_dict_default()

        if self.dump_only:
            return

        self.allow_stdin_stdout = True
        self.unlink = bool(unlink)

        template_path = template_path or J2SUBST_TEMPLATE_PATH_PARTS
        self.template_path: list[str] = non_empty_str(template_path)

        self.dict_env: dict[str, str] = {}
        self.j2fs_loaders: dict[str, jinja2.FileSystemLoader] = {}

        self.resolve_template_path(resolve_placeholders=False)

        ## make shallow copy of os.environ (for good)
        # self.dict_env = os.environ
        self.dict_env = { k: v for k, v in os.environ.items() if not is_env_skipped(k) }

        j2ext = list(J2SUBST_JINJA_EXTENSIONS)
        if self.debug:
            j2ext += J2SUBST_JINJA_DEBUG_EXTENSIONS

        self.j2env = jinja2.Environment(
            extensions=j2ext,
            ## dumb loader: does nothing by default
            loader=jinja2.DictLoader( { } ),
        )

        for m in J2SUBST_PYTHON_MODULES:
            self.import_python_module(m)
        for alias, m in J2SUBST_PYTHON_MODULE_ALIASES.items():
            self.import_python_module(m, alias)

        def __not_valid_python_module(m: Any):
            raise ValueError(f'not valid "python_modules": {repr(m)}')

        if python_modules is None:
            pass
        elif is_seq(python_modules):
            for m in python_modules:
                if not isinstance(m, str):
                    __not_valid_python_module(m)
                self.import_python_module(m)
        elif is_map(python_modules):
            for alias, m in python_modules.items():
                if not (isinstance(alias, str) and isinstance(m, str)):
                    __not_valid_python_module( {alias: m} )
                self.import_python_module(m, alias)
        else:
            __not_valid_python_module(python_modules)

        for f in J2SUBST_BUILTIN_FUNCTIONS:
            self.import_builtin_function(f)
        for alias, f in J2SUBST_BUILTIN_FUNCTION_ALIASES.items():
            self.import_builtin_function(f, alias)

        for f in J2SUBST_FILTERS:
            self.import_filter(f)
        for alias, f in J2SUBST_FILTER_ALIASES.items():
            self.import_filter(f, alias)

        if filters_as_functions:
            for f in J2SUBST_FILTERS:
                self.import_function(f)
            for alias, f in J2SUBST_FILTER_ALIASES.items():
                self.import_function(f, alias)

    def __verify_dump_only(self):
        if not self.dump_only:
            return
        raise ValueError('"dump_only" is True')

    def __warn(self, source: str, message: str):
        if self.strict:
            raise ValueError(message)
        if (self.verbosity >= 0) or self.debug:
            print(f'J2subst: {source}: {message}', file=sys.stderr)

    def __info(self, source: str, message: str):
        if (self.verbosity > 0) or self.debug:
            print(f'J2subst: {source}: {message}', file=sys.stderr)

    def __debug(self, source: str, message: str):
        if self.debug:
            print(f'J2subst: {source}: {message}', file=sys.stderr)

    def __merge_dict_default(self):

        def _warn(msg: str):
            self.__warn('merge_dict_default', msg)

        def _info(msg: str):
            self.__info('merge_dict_default', msg)

        for p in self.config_path:
            if os.path.isfile(p):
                if not self.merge_dict_from_file(p):
                    _warn(f'failed to load config file: {p}')
            elif os.path.isdir(p):
                _entries: list[str] = []
                for e in os.listdir(p):
                    if e.startswith('.'):
                        ## silently ignore hidden files
                        continue
                    ext = os.path.splitext(e)[1]
                    if (not ext) or ext not in J2SUBST_CONFIG_EXT:
                        ## silently ignore non-recognized extensions
                        continue
                    _entries.append(e)

                for e in natsort.natsorted(_entries):
                    f = os.path.join(p, e)
                    if not os.path.isfile(f):
                        continue

                    real_f = os.path.realpath(f)
                    if f == real_f:
                        _info(f'try loading {f}')
                    else:
                        _info(f'try loading {f} <- {real_f}')

                    if not self.merge_dict_from_file(f):
                        _warn(f'failed to load config file: {f}')
            else:
                _warn(f'not a file or directory, or does not exist: {p}')

    def __ensure_fs_loader_for(self, path: str | PathLike[str]) -> bool:
        if not os.path.isdir(path):
            return False

        p = str(path)
        if p in self.j2fs_loaders:
            return True

        l = jinja2.FileSystemLoader(
            path, encoding='utf-8', followlinks=True,
        )
        self.j2fs_loaders.update( { p: l } )

        return True

    def ensure_fs_loader_for(self, path: str | PathLike[str]) -> bool:
        self.__verify_dump_only()

        def __warn(msg: str):
            self.__warn('ensure_fs_loader_for', msg)

        d = os.path.abspath(path)
        if os.path.isdir(d):
            pass
        elif os.path.isfile(d):
            d = os.path.dirname(d)
        else:
            __warn(f'not a file or directory, or does not exist: {repr(path)}')
            return False

        return self.__ensure_fs_loader_for(d)

    def __resolve_origin(self, origin: str | PathLike[str] | None = None) -> tuple[str | None, bool]:

        def __warn(msg: str):
            self.__warn('__resolve_origin', msg)

        if origin is None:
            return (None, False)

        if not os.path.exists(origin):
            __warn(f'does not exist: {repr(origin)}')
            return (None, False)

        _origin = os.path.normpath(origin)
        if os.path.isdir(_origin):
            _origin = str(_origin)
        elif os.path.isfile(_origin):
            _origin = str(os.path.dirname(_origin))
            if _origin == '':
                _origin = '.'
        else:
            __warn(f'not a file or directory: {repr(origin)}')
            return (None, False)

        _want_root = _origin.startswith('/')
        _origin = str(os.path.abspath(_origin))

        return (_origin, _want_root)

    def resolve_template_path(self, resolve_placeholders: bool, origin: str | PathLike[str] | None = None) -> list[str]:
        self.__verify_dump_only()

        def __warn(msg: str):
            self.__warn('resolve_template_path', msg)

        def __info(msg: str):
            self.__info('resolve_template_path', msg)

        def __debug(msg: str):
            self.__debug('resolve_template_path', msg)

        _origin, _want_root = self.__resolve_origin(origin)

        # dirs: list[str | PathLike[str]] = []
        dirs: list[str] = []

        for p in self.template_path:
            ## see J2SUBST_TEMPLATE_PATH_PARTS
            s = str(p)
            if s.find('@{') >= 0:
                if not resolve_placeholders:
                    __debug(f'not going to resolve template path: {repr(s)}')
                    continue

                s = s.replace('@{CWD}', os.getcwd())
                if _origin is not None:
                    s = s.replace('@{ORIGIN}', _origin)

                ## skip if there're still some special placeholders
                if s.find('@{') >= 0:
                    __warn(f'failed to resolve template path: {repr(s)}')
                    continue

            s = str(os.path.abspath(os.path.normpath(s)))
            if s in dirs:
                __info(f'duplicate template path: {repr(s)}')
                continue

            dirs.append(s)

        if _want_root:
            if '/' in dirs:
                __info('duplicate template path: "/"')
            else:
                dirs.append('/')

        dirs_final: list[str] = []
        for d in dirs:
            if self.__ensure_fs_loader_for(d):
                dirs_final.append(d)
                continue
            __warn(f'failed to load template directory: {repr(d)}')

        # return [str(d) for d in dirs_final]
        return dirs_final

    def remove_global(self, name: str):
        self.__verify_dump_only()

        self.j2env.globals.pop(name, None)

    def remove_filter(self, name: str):
        self.__verify_dump_only()

        self.j2env.filters.pop(name, None)

    def __import_python_module(self, module_name: str, alias: str):
        self.j2env.globals.update( { alias: importlib.import_module(module_name) } )

    def import_python_module(self, module_name: str, alias: str | None = None):
        self.__verify_dump_only()

        def __warn(msg: str):
            self.__warn('import_python_module', msg)

        n = alias or module_name
        if not is_plain_key(n):
            __warn(f'key is not "plain", module {repr(module_name)} will not be imported as {repr(n)}')
            return
        if n in self.j2env.globals:
            __warn(f'globals already has {repr(n)} key, module {repr(module_name)} will not be imported as {repr(n)}')
            return
        self.__import_python_module(module_name, n)

    def __import_filter(self, func: Any, alias: str):
        self.j2env.filters.update( { alias: func } )

    def import_filter(self, func: Any, alias: str | None = None):
        self.__verify_dump_only()

        if not callable(func):
            raise ValueError('func is not callable')

        def __warn(msg: str):
            self.__warn('import_filter', msg)

        n = alias or func.__name__
        if not is_plain_key(n):
            __warn(f'key is not "plain", filter {repr(func.__name__)} will not be imported as {repr(n)}')
            return
        if n in self.j2env.filters:
            __warn(f'filters already has {repr(n)} key, filter {repr(func.__name__)} will not be imported as {repr(n)}')
            return
        self.__import_filter(func, n)

    def __import_function(self, func: Any, alias: str):
        self.j2env.globals.update( { alias: func } )

    def import_builtin_function(self, func: Any, alias: str | None = None):
        self.__verify_dump_only()

        if not callable(func):
            raise ValueError('func is not callable')

        def __warn(msg: str):
            self.__warn('import_builtin_function', msg)

        n = alias or func.__name__
        if not is_plain_key(n):
            __warn(f'key is not "plain", builtin function {repr(func.__name__)} will not be imported as {repr(n)}')
            return
        if n in self.j2env.globals:
            __warn(f'globals already has {repr(n)} key, builtin function {repr(func.__name__)} will not be imported as {repr(n)}')
            return
        self.__import_function(func, n)

    def import_function(self, func: Any, alias: str | None = None):
        self.__verify_dump_only()

        if not callable(func):
            raise ValueError('func is not callable')

        def __warn(msg: str):
            self.__warn('import_function', msg)

        n = alias or func.__name__
        if not is_plain_key(n):
            __warn(f'key is not "plain", function {repr(func.__name__)} will not be imported as {repr(n)}')
            return
        if n in self.j2env.globals:
            __warn(f'globals already has {repr(n)} key, function {repr(func.__name__)} will not be imported as {repr(n)}')
        self.__import_function(func, n)

    def merge_dict_from_yaml(self, filename: str | PathLike[str]):
        yaml_all_empty = True
        with open(filename, mode='r', encoding='utf-8') as fx:
            for x in yaml.safe_load_all(fx):
                if not x:
                    continue
                yaml_all_empty = False
                self.dict_cfg = merge_dict_recurse(self.dict_cfg, x)

        if yaml_all_empty:
            self.__info('merge_dict_from_yaml', f'received empty document(s) from: {repr(filename)}')

    def merge_dict_from_toml(self, filename: str | PathLike[str]):
        with open(filename, mode='rb') as fx:
            x = tomllib.load(fx)
            self.dict_cfg = merge_dict_recurse(self.dict_cfg, x)

    def merge_dict_from_json(self, filename: str | PathLike[str]):
        with open(filename, mode='r', encoding='utf-8') as fx:
            x = json.load(fx)
            self.dict_cfg = merge_dict_recurse(self.dict_cfg, x)

    def merge_dict_from_json_str(self, string: str):
        x = json.loads(string)
        self.dict_cfg = merge_dict_recurse(self.dict_cfg, x)

    def merge_dict_from_file(self, filename: str | PathLike[str]) -> bool:
        if not filename:
            return False

        def __warn(msg: str):
            self.__warn('merge_dict_from_file', msg)

        if not os.path.isfile(filename):
            __warn(f'not a file or does not exist: {repr(filename)}')
            return False

        ext = os.path.splitext(filename)[1]
        if ext not in J2SUBST_CONFIG_EXT:
            __warn(f'non-recognized name extension: {repr(filename)}')
            return False

        if ext in [ '.yml', '.yaml' ]:
            self.merge_dict_from_yaml(filename)
        elif ext == '.toml':
            self.merge_dict_from_toml(filename)
        elif ext == '.json':
            self.merge_dict_from_json(filename)
        else:
            ## likely unreachable
            __warn(f'non-recognized name extension: {repr(filename)}')
            return False

        return True

    def dump_config_yaml(self) -> str:
        if not self.dict_cfg:
            return J2SUBST_EMPTY_YAML
        return yaml.safe_dump(self.dict_cfg, sort_keys=True)

    def dump_config_json(self) -> str:
        if not self.dict_cfg:
            return J2SUBST_EMPTY_JSON
        return json.dumps(self.dict_cfg, sort_keys=True)

    def dump_config(self, fmt: J2substDumpFormat = J2SUBST_DUMP_FORMAT) -> str:
        if fmt == J2substDumpFormat.YAML:
            return self.dump_config_yaml()
        if fmt == J2substDumpFormat.JSON:
            return self.dump_config_json()
        raise ValueError(f'unknown dump format: {repr(fmt)}')

    def env_overlay(self, j2subst_origin: str | PathLike[str] | None = None, **kwargs: dict[str, Any]) -> jinja2.Environment:
        self.__verify_dump_only()

        kw: dict[str, Any] = {}
        if kwargs:
            kw = kw | kwargs

        _x = kw.get('loader')
        if (_x is not None) and isinstance(_x, jinja2.BaseLoader):
            pass
        else:
            dirs: list[str] = self.resolve_template_path(resolve_placeholders=True, origin=j2subst_origin)

            loader: jinja2.BaseLoader
            if dirs:
                loader = jinja2.ChoiceLoader(
                    [ self.j2fs_loaders[d] for d in dirs ]
                )
            else:
                loader=jinja2.DictLoader( { } )

            kw.update( { 'loader': loader } )

        return self.j2env.overlay(**kw)

    def __prepare_kwargs(self, j2subst_file: str | None, j2subst_origin: str | None) -> dict[str, Any]:
        kw: dict[str, Any] = {
            self.dict_cfg_name: self.dict_cfg,
            self.dict_env_name: self.dict_env,
        }
        kw.update( {
            ## hardcoded:
            'is_ci': is_ci(),
            'j2subst_file': j2subst_file,
            'j2subst_origin': j2subst_origin,
        } )

        return kw

    def render_str(self, string: str, j2env_overlay: jinja2.Environment | None = None) -> tuple[str, str | None]:
        self.__verify_dump_only()

        _env = j2env_overlay
        if _env is None:
            _env = self.env_overlay()
        t = _env.from_string(string)

        kw = self.__prepare_kwargs(None, None)

        return t.render(**kw), None

    def render_text_io(self, io_source: io.TextIOBase, j2env_overlay: jinja2.Environment | None = None) -> tuple[str, str | None]:
        return self.render_str(''.join(io_source.readlines()), j2env_overlay)

    def render_from_file(self, filename: str, j2env_overlay: jinja2.Environment | None = None) -> tuple[str, str | None]:
        self.__verify_dump_only()

        def __debug(msg: str):
            self.__debug('render_from_file', msg)

        _env = j2env_overlay
        if _env is None:
            ## preserve internal settings
            (_v, _d, _s) = (self.verbosity, self.debug, self.strict)
            ## override internal settings for this call
            (self.verbosity, self.debug, self.strict) = (-1, False, False)
            _env = self.env_overlay()
            ## restore internal settings
            (self.verbosity, self.debug, self.strict) = (_v, _d, _s)

            __debug('trying to resolve with self.env_overlay()')

            ## TODO: avoid try-except
            try:
                _env.get_template(filename)
            except jinja2.TemplateNotFound:
                __debug(f'jinja2.TemplateNotFound: {repr(filename)}')
                __debug(f'trying to resolve with self.env_overlay({repr(filename)})')

                _env = self.env_overlay(filename)

        t = _env.get_template(filename)
        _origin, _ = self.__resolve_origin(t.filename)

        kw = self.__prepare_kwargs(t.filename, _origin)

        return t.render(**kw), t.filename

    def render_stdin(self, j2env_overlay: jinja2.Environment | None = None) -> str:
        self.__verify_dump_only()

        r, _ = self.render_text_io(sys.stdin, j2env_overlay)
        return r

    def render_file(self, file_in: str | PathLike[str], file_out: str | PathLike[str] | None = None, j2env_overlay: jinja2.Environment | None = None) -> bool:
        self.__verify_dump_only()

        rendered: str
        f_in: str | None = None
        f_stdin: bool = False
        f_out: str | None = None

        def __warn(msg: str):
            self.__warn('render_file', msg)

        def __render_error(msg: str) -> bool:
            __warn(msg)
            return False

        def __info(msg: str):
            self.__info('render_file', msg)

        def __debug(msg: str):
            self.__debug('render_file', msg)

        if is_stdin(file_in):
            if not self.allow_stdin_stdout:
                return __render_error('stdin not allowed')
            f_stdin = True
            rendered = self.render_stdin(j2env_overlay)
        else:
            rendered, f_in = self.render_from_file(str(file_in), j2env_overlay)

        if file_out is None:
            if f_stdin:
                f_out = '-'
            else:
                if f_in is None:
                    return __render_error('unable to determine output file name')
                if not f_in.endswith(J2SUBST_TEMPLATE_EXT):
                    return __render_error(f'input file name extension mismatch: {repr(file_in)}')
                f_out = os.path.splitext(f_in)[0]
        else:
            f_out = str(file_out)

        ## safety measures
        if f_out is None:
            return __render_error('unable to determine output file name')

        if is_stdout(f_out):
            if not self.allow_stdin_stdout:
                return __render_error('stdout not allowed')

            sys.stdout.write(rendered)
            sys.stdout.flush()

            if self.unlink:
                if f_stdin:
                    __info('cannot unlink() stdin')
                elif f_in:
                    os.unlink(f_in)

            return True

        ## TODO: there're still TOCTOU windows

        ## safety measures
        if os.path.islink(f_out):
            return __render_error(f'output file is symlink: {f_out}')
        if os.path.exists(f_out):
            if not os.path.isfile(f_out):
                return __render_error(f'output file is not a file: {f_out}')
            if f_in and os.path.samefile(f_in, f_out):
                return __render_error(f'unable to process template inplace: {f_in}')
            if not self.force:
                return __render_error(f'unable to overwrite existing file: {f_out}')

            os.unlink(f_out)

        with open(f_out, mode='w', encoding='utf-8') as f:
            f.reconfigure(write_through=True)
            f.write(rendered)
            f.flush()

        if self.unlink:
            if f_stdin:
                __info('cannot unlink() stdin')
            elif f_in:
                os.unlink(f_in)

        return True

    def render_directory(self, directory: str | PathLike[str], depth: int = 1, j2env_overlay: jinja2.Environment | None = None) -> bool:
        self.__verify_dump_only()

        def __warn(msg: str):
            self.__warn('render_directory', msg)

        def __render_error(msg: str) -> bool:
            __warn(msg)
            return False

        def __info(msg: str):
            self.__info('render_directory', msg)

        def __debug(msg: str):
            self.__debug('render_directory', msg)

        if not os.path.isdir(directory):
            return __render_error(f'not a directory: {repr(directory)}')

        ## minor adjustments
        if depth < 0:
            depth = -1
        if depth == 0:
            __debug('depth == 0')
            return True

        rv = True

        _entries: list[str] = []
        for e in os.listdir(directory):
            if e.startswith('.'):
                ## silently ignore hidden files
                continue
            _entries.append(e)

        for e in natsort.natsorted(_entries):
            p = os.path.join(directory, e)

            if os.path.isdir(p):
                if depth < 0:
                    rv &= self.render_directory(p, depth, j2env_overlay)
                else:
                    rv &= self.render_directory(p, depth - 1, j2env_overlay)
                continue

            if e.endswith(J2SUBST_TEMPLATE_EXT) and os.path.isfile(p):
                rv &= self.render_file(p, None, j2env_overlay)
                continue

            __info(f'ignore: {e}')

        return rv

    def handle_simple_cli_args(self, arg1: str | PathLike[str], arg2: str | PathLike[str] | None = None) -> tuple[str | None, str | None]:
        _in, _out = (None, None)

        ## 1st argument is missing or empty
        if not arg1:
            return (_in, _out)

        if is_stdin(arg1):
            _in = '-'
        else:
            a = os.path.normpath(arg1)
            if os.path.isfile(a) and a.endswith(J2SUBST_TEMPLATE_EXT):
                _in = str(arg1)

        ## early exit:
        ## 1) 1st argument not a stdin or file with supported file name extension
        ## 2) 2nd argument is missing or empty
        if (_in is None) or (not arg2):
            return (_in, _out)

        if is_stdout(arg2):
            _out = '-'
        else:
            a = os.path.normpath(arg2)
            if a.endswith(J2SUBST_TEMPLATE_EXT):
                ## perform regular processing: 2nd argument is (probably) template too
                _in = None
            else:
                _out = a

        return (_in, _out)



#!/usr/bin/env python3

import os

from typing import (
    Any,
)

## click
import click
import click.core

## this module
from .dumpfmt import J2substDumpFormat
from .defaults import (
    J2SUBST_DICT_NAME_CFG,
    J2SUBST_DICT_NAME_ENV,
    J2SUBST_DUMP_FORMAT,
    J2SUBST_MAX_DEPTH,
    J2SUBST_TEMPLATE_PATH_PARTS,
    J2SUBST_TEMPLATE_PATH,
    J2SUBST_VERSION,
)
from .functions import (
    click_bool,
    click_bool_neg,
    is_ci,
    is_str,
    str_split_to_list,
)
from .cli_help import (
    J2SUBST_CLI_HELP__CICD,
    J2SUBST_CLI_HELP__CLICK,
    J2SUBST_CLI_HELP__DUMP,
    J2SUBST_CLI_HELP__ENV,
    J2SUBST_CLI_HELP__TEMPLATE_PATH,
)
from .j2subst import J2subst


## NB: click.option() with "show_envvar=True" does a somewhat horrible formatting
## so we do it manually (see cli_help.py)

J2SUBST_CLI_HELP_CONFIG_PATH = '''
    Colon-separated list of configuration files or directories.
    Supported formats: YAML (".yaml", ".yml"), JSON (".json"), TOML (".toml").
'''

J2SUBST_CLI_HELP_TEMPLATE_PATH = f'''
    Colon-separated list of template directories.

    Default: {J2SUBST_TEMPLATE_PATH}
'''

J2SUBST_CLI_HELP_PYTHON_MODULES = '''
    Space-separated list of Python modules to import.

    To import module with an alias, use format: <alias_name>:<module_name>.
'''


def __dump_callback(_ctx: Any, _param: Any, value: str | bool | None) -> J2substDumpFormat | None:
    if value is None:
        return None

    if value is True:
        return J2SUBST_DUMP_FORMAT
    if value is False:
        return None

    if is_str(value):
        if click_bool(value):
            return J2SUBST_DUMP_FORMAT
        if click_bool_neg(value):
            return None
        value = value.upper()
        if value in [ e.name for e in J2substDumpFormat ]:
            return J2substDumpFormat[value]

    raise click.BadParameter(f"must be one of: {', '.join([e.name for e in J2substDumpFormat])}.")


J2SUBST_CLI_CONTEXT_SETTINGS = {
    'help_option_names': ['-h', '--help'],
}


@click.command(
    no_args_is_help=not is_ci(),
    context_settings=J2SUBST_CLI_CONTEXT_SETTINGS,
)
@click.version_option(version=J2SUBST_VERSION,
    prog_name='j2subst',
    message='%(prog)s %(version)s'
)

## extra help topics
@click.option('--help-cicd',
    'o_help_cicd', is_flag=True,
    help='Show help for J2subst behavior in CI/CD.',
)
@click.option('--help-click',
    'o_help_click', is_flag=True,
    help='Show help for Click behavior.',
)
@click.option('--help-dump',
    'o_help_dump', is_flag=True,
    help='Show help for "--dump" mode.',
)
@click.option('--help-env',
    'o_help_env', is_flag=True,
    help='Show help for environment variables.',
)
@click.option('--help-template-path',
    'o_help_template_path', is_flag=True,
    help='Show help for "--template-path".',
)

@click.option('--dump',
    'o_dump_fmt', is_flag=False, flag_value=True, type=click.UNPROCESSED, callback=__dump_callback,
    help='Dump configuration to stdout and exit.',
    metavar='FORMAT',
)

@click.option('--verbose', '-v',
    'o_verbose', count=True,
    envvar='J2SUBST_VERBOSE',
    help='Increase verbosity.',
)
@click.option('--quiet', '-q',
    'o_quiet', is_flag=True,
    envvar='J2SUBST_QUIET',
    help='Enable quiet mode (overrides any --verbose).',
)

@click.option('--debug', '-D',
    'o_debug', is_flag=True,
    envvar='J2SUBST_DEBUG',
    help='Enable debug mode.',
)
@click.option('--strict', '-s',
    'o_strict', is_flag=True,
    envvar='J2SUBST_STRICT',
    help='Enable strict mode.',
)
@click.option('--force', '-f',
    'o_force', is_flag=True,
    envvar='J2SUBST_FORCE',
    help='Enable force mode.',
)
@click.option('--unlink', '-u',
    'o_unlink', is_flag=True,
    envvar='J2SUBST_UNLINK',
    help='Delete template files after expanding it.',
)
@click.option('--depth', '-d',
    'o_depth', type=click.IntRange(1, J2SUBST_MAX_DEPTH),
    envvar='J2SUBST_DEPTH',
    help='Set recursion depth to look for template files.',
    metavar='INTEGER',
)

@click.option('--config-path', '-c',
    'o_config_path',
    envvar='J2SUBST_CONFIG_PATH',
    help=J2SUBST_CLI_HELP_CONFIG_PATH,
    metavar='LIST',
)
@click.option('--template-path', '-t',
    'o_template_path',
    default=J2SUBST_TEMPLATE_PATH,
    envvar='J2SUBST_TEMPLATE_PATH',
    help=J2SUBST_CLI_HELP_TEMPLATE_PATH,
    metavar='LIST',
)

## extra options
@click.option('--python-modules',
    'o_python_modules',
    envvar='J2SUBST_PYTHON_MODULES',
    help=J2SUBST_CLI_HELP_PYTHON_MODULES,
    metavar='LIST',
)
@click.option('--fn-filters',
    'o_fn_filters', is_flag=True,
    envvar='J2SUBST_FN_FILTERS',
    help='Propagate filters as functions too.',
)
@click.option('--dict-name-cfg',
    'o_dict_name_cfg',
    envvar='J2SUBST_DICT_NAME_CFG',
    help='Assign configuration dictionary name to custom value.',
)
@click.option('--dict-name-env',
    'o_dict_name_env',
    envvar='J2SUBST_DICT_NAME_ENV',
    help='Assign environment dictionary name to custom value.',
)

@click.argument('cli_args',
    nargs=-1,
    metavar='[ARGUMENTS]',
)

@click.pass_context
def cli(ctx: click.Context,

        ## extra help topics
        o_help_cicd: bool,
        o_help_click: bool,
        o_help_dump: bool,
        o_help_env: bool,
        o_help_template_path: bool,

        o_dump_fmt: J2substDumpFormat | None,

        o_verbose: int,
        o_quiet: bool,
        o_debug: bool,
        o_strict: bool,
        o_force: bool,
        o_unlink: bool,
        o_depth: int | None,
        o_config_path: str | None,
        o_template_path: str | None,

        o_python_modules: str | None,
        o_fn_filters: bool,
        o_dict_name_cfg: str | None,
        o_dict_name_env: str | None,

        cli_args: tuple[str],
):

    ## extra help topics
    help_topics = [
        [ o_help_cicd,          J2SUBST_CLI_HELP__CICD, ],
        [ o_help_click,         J2SUBST_CLI_HELP__CLICK, ],
        [ o_help_dump,          J2SUBST_CLI_HELP__DUMP, ],
        [ o_help_env,           J2SUBST_CLI_HELP__ENV, ],
        [ o_help_template_path, J2SUBST_CLI_HELP__TEMPLATE_PATH, ],
    ]
    for h in help_topics:
        if h[0]:
            click.echo(h[1])
            ctx.exit()

    ## verify command-line usage
    if o_dump_fmt is not None:

        def __from_cmdline(key: str):
            return ctx.get_parameter_source(key) == click.core.ParameterSource.COMMANDLINE

        def __dump_usage_error(key: str, flag: str):
            if __from_cmdline(key):
                raise click.UsageError(f'Cannot use --dump with {flag}', ctx)

        __dump_usage_error('o_force',  '--force')
        __dump_usage_error('o_unlink', '--unlink')
        __dump_usage_error('o_depth',  '--depth')

        __dump_usage_error('o_template_path', '--template-path')

        __dump_usage_error('o_python_modules', '--python-modules')
        __dump_usage_error('o_fn_filters',     '--fn-filters')
        __dump_usage_error('o_dict_name_cfg',  '--dict-name-cfg')
        __dump_usage_error('o_dict_name_env',  '--dict-name-env')

        ## TODO: support --dump with output file name
        if len(list(cli_args)) > 0:
            raise click.UsageError("Cannot use --dump with arguments", ctx)

    ## adjust verbosity level:
    ## -1 - quiet
    ##  0 - warnings
    ##  1 - info
    ##  2 and more - not implemented
    if o_quiet:
        o_verbose = -1

    _config_path = None
    if o_config_path is not None:
        _config_path = str_split_to_list(o_config_path, ':')

    if o_dump_fmt is not None:
        j = J2subst(dump_only=True,
            verbosity=o_verbose,
            debug=o_debug,
            strict=o_strict,
            config_path=_config_path,
        )

        ## TODO: support --dump with output file name
        print(j.dump_config(o_dump_fmt), flush=True)
        ctx.exit()

    ## here goes regular processing

    if o_depth is None:
        if is_ci():
            o_depth = J2SUBST_MAX_DEPTH
        else:
            o_depth = 1

    if o_template_path is None:
        _template_path = J2SUBST_TEMPLATE_PATH_PARTS
    else:
        _template_path = str_split_to_list(o_template_path, ':')

    if o_dict_name_cfg is None:
        o_dict_name_cfg = J2SUBST_DICT_NAME_CFG
    if o_dict_name_env is None:
        o_dict_name_env = J2SUBST_DICT_NAME_ENV

    _python_modules: dict[str,str] = {}
    if o_python_modules is not None:
        _modules = str_split_to_list(o_python_modules)
        for m in _modules:
            # if m.index(':') < 0:
            #     _python_modules.update( { m: m, } )
            #     continue
            _m = str_split_to_list(m, ':')
            _len = len(_m)
            if _len == 1:
                _python_modules.update( { _m[0]: _m[0], } )
            elif _len == 2:
                _python_modules.update( { _m[0]: _m[1], } )
            else:
                raise click.UsageError(f'not valid "python_modules": {repr(m)}', ctx)

    args: list[str] = list(cli_args)
    if len(args) == 0:
        if is_ci():
            args = [ os.getcwd() ]
        else:
            raise click.UsageError('no arguments were specified', ctx)

    j = J2subst(
            verbosity=o_verbose,
            debug=o_debug,
            strict=o_strict,
            force=o_force,
            unlink=o_unlink,

            config_path=_config_path,
            template_path=_template_path,

            python_modules=_python_modules,
            filters_as_functions=o_fn_filters,
            dict_name_cfg=o_dict_name_cfg,
            dict_name_env=o_dict_name_env,
    )

    ## deal with 1/2 argument mode
    _in, _out = j.handle_simple_cli_args(*args[:2])

    r = True
    if _in:
        r &= j.render_file(_in, _out)
    else:
        ## disallow stdin/stdout from this moment
        j.allow_stdin_stdout = False

        for arg in args:
            if os.path.isdir(arg):
                r &= j.render_directory(arg, o_depth)
            else:
                r &= j.render_file(arg)

    if not r:
        ctx.exit(1)

    ctx.exit(0)
    return


if __name__ == '__main__':
    # pylint: disable=E1120
    cli(
        prog_name='j2subst',
    )
